import os

USER_HOME_ELEANOR_CACHE = os.path.join(os.path.expanduser('~'), '.eleanor/')
USER_HOME_SHERLOCKPIPE = os.path.join(os.path.expanduser('~'), '.eleanor/')
AU_TO_RSUN = 215.032
MOMENTUM_DUMP_QUALITY_FLAG = 2 ** 5